# Generated by Django 2.2.28 on 2023-07-07 15:51
from django.db import IntegrityError, migrations
from django.utils.text import slugify

from sentry.constants import ObjectStatus
from sentry.new_migrations.migrations import CheckedMigration
from sentry.utils.query import RangeQuerySetWrapperWithProgressBar


def schedule(cls, instance, days=30):
    from datetime import timedelta

    from django.utils import timezone

    model = type(instance)
    model_name = model.__name__
    cls.objects.update_or_create(
        app_label=instance._meta.app_label,
        model_name=model_name,
        object_id=instance.pk,
        defaults={
            "actor_id": None,
            "data": {},
            "date_scheduled": timezone.now() + timedelta(days=days, hours=0),
        },
    )


def migrate_monitor_slugs(apps, schema_editor):
    Monitor = apps.get_model("sentry", "Monitor")
    Rule = apps.get_model("sentry", "Rule")
    RegionScheduledDeletion = apps.get_model("sentry", "RegionScheduledDeletion")
    ScheduledDeletion = apps.get_model("sentry", "ScheduledDeletion")

    MAX_SLUG_LENGTH = 50

    for monitor in RangeQuerySetWrapperWithProgressBar(Monitor.objects.all()):
        monitor_slug = monitor.slug
        slugified = slugify(monitor_slug)[:MAX_SLUG_LENGTH].strip("-")

        # Nothing to migrate if the monitor already has a valid slug
        if monitor_slug == slugified:
            continue

        try:
            monitor.slug = slugified
            monitor.save()
        except IntegrityError:
            # If there is a collision, delete the old monitor as the new one is receiving all check-ins
            alert_rule_id = monitor.config.get("alert_rule_id")
            if alert_rule_id:
                rule = (
                    Rule.objects.filter(
                        project_id=monitor.project_id,
                        id=alert_rule_id,
                    )
                    .exclude(
                        status__in=[
                            ObjectStatus.PENDING_DELETION,
                            ObjectStatus.DELETION_IN_PROGRESS,
                        ]
                    )
                    .first()
                )
                if rule:
                    rule.status = ObjectStatus.PENDING_DELETION
                    rule.save()
                    schedule(RegionScheduledDeletion, rule, days=0)

            # Revert slug so as not to attempt re-saving clean slug
            monitor.slug = monitor_slug
            monitor.status = ObjectStatus.PENDING_DELETION
            monitor.save()
            schedule(ScheduledDeletion, monitor, days=0)


class Migration(CheckedMigration):
    # This flag is used to mark that a migration shouldn't be automatically run in production. For
    # the most part, this should only be used for operations where it's safe to run the migration
    # after your code has deployed. So this should not be used for most operations that alter the
    # schema of a table.
    # Here are some things that make sense to mark as dangerous:
    # - Large data migrations. Typically we want these to be run manually by ops so that they can
    #   be monitored and not block the deploy for a long period of time while they run.
    # - Adding indexes to large tables. Since this can take a long time, we'd generally prefer to
    #   have ops run this and not block the deploy. Note that while adding an index is a schema
    #   change, it's completely safe to run the operation after the code has deployed.
    is_dangerous = False

    dependencies = [
        ("sentry", "0514_migrate_priority_saved_searches"),
    ]

    operations = [
        migrations.RunPython(
            migrate_monitor_slugs,
            migrations.RunPython.noop,
            hints={"tables": ["sentry_monitor"]},
        ),
    ]
